package obuilder

  1. Overview
  2. Docs

Module Obuilder.Store_specSource

Configuration information to set up a store.

Sourcetype t = [
  1. | `Btrfs of string
  2. | `Zfs of string
  3. | `Rsync of string * Rsync_store.mode
  4. | `Xfs of string
  5. | `Overlayfs of string
  6. | `Docker of string
  7. | `Qemu of string
  8. | `Hcs of string
]
Sourceval is_absolute : string -> bool
Sourceval of_string : string -> ([> `Btrfs of string | `Docker of string | `Hcs of string | `Overlayfs of string | `Qemu of string | `Rsync of string | `Xfs of string | `Zfs of string ], [> `Msg of string ]) result
Sourceval pp : Format.formatter -> [< `Btrfs of string | `Docker of string | `Hcs of string | `Overlayfs of string | `Qemu of string | `Rsync of string | `Xfs of string | `Zfs of string ] -> unit
Sourcetype store =
  1. | Store : (module S.STORE with type t = 'a) * 'a -> store
Sourceval to_store : [< `Btrfs of string | `Docker of string | `Hcs of string | `Overlayfs of string | `Qemu of string | `Rsync of string * Rsync_store.mode | `Xfs of string | `Zfs of string ] -> [> `Docker | `Hcs | `Native | `Qemu ] * store Lwt.t
Sourceval store_t : [ `Btrfs of string | `Docker of string | `Hcs of string | `Overlayfs of string | `Qemu of string | `Rsync of string | `Xfs of string | `Zfs of string ] Cmdliner.Arg.conv
Sourceval store : ?docs:??? -> string list -> [ `Btrfs of string | `Docker of string | `Hcs of string | `Overlayfs of string | `Qemu of string | `Rsync of string | `Xfs of string | `Zfs of string ] option Cmdliner.Arg.t
Sourceval rsync_mode_opt : Rsync_store.mode option Cmdliner.Arg.t
Sourceval rsync_mode : Rsync_store.mode option Cmdliner.Term.t
Sourceval of_t : [> `Btrfs of 'a | `Docker of 'b | `Hcs of 'c | `Overlayfs of 'd | `Qemu of 'e | `Rsync of 'f | `Xfs of 'g | `Zfs of 'h ] option -> 'i option -> [> `Btrfs of 'a | `Docker of 'b | `Hcs of 'c | `Overlayfs of 'd | `Qemu of 'e | `Rsync of 'f * 'i | `Xfs of 'g | `Zfs of 'h ]

Transform a store and rsync-mode into a validated combination.

For example an rsync store must supply an rsync-mode.

Sourceval v : [> `Btrfs of string | `Docker of string | `Hcs of string | `Overlayfs of string | `Qemu of string | `Rsync of string * Rsync_store.mode | `Xfs of string | `Zfs of string ] Cmdliner.Term.t

Parse cli arguments for t

Sourceval cmdliner : ([> `Docker | `Hcs | `Native | `Qemu ] * store Lwt.t) Cmdliner.Term.t

Parse cli arguments for t and initialise a store.